// Mouse filter from ICE, refactored a bit

#include <stdio.h>
#ifdef WIN32 
	#define NOMINMAX
	#include <windows.h>
#elif LINUX
#endif
#include "Nx.h"
#include "MouseFilter.h"

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/**
 *	Constructor.
 */
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
MouseFilter::MouseFilter()
{
	mWeightModifier			= 0.0f;
	mHistoryBufferLength	= 0;
	mHistoryBufferX			= NULL;
	mHistoryBufferY			= NULL;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/**
 *	Constructor.
 */
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
MouseFilter::MouseFilter(NxU32 length, NxF32 weight_modifier)
{
	mWeightModifier			= weight_modifier;
	mHistoryBufferLength	= 0;
	mHistoryBufferX			= NULL;
	mHistoryBufferY			= NULL;
	SetHistoryBufferLength(length);
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/**
 *	Destructor.
 */
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
MouseFilter::~MouseFilter()
{
	NX_DELETE_ARRAY(mHistoryBufferX);
	NX_DELETE_ARRAY(mHistoryBufferY);
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/**
 *	Sets the length of the history buffer.
 *	\param		length	[in] buffer length
 *	\return		true if success
 */
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool MouseFilter::SetHistoryBufferLength(NxU32 length)
{
	NX_DELETE_ARRAY(mHistoryBufferX);
	NX_DELETE_ARRAY(mHistoryBufferY);

	mHistoryBufferLength = length;
	if(length)
	{
		mHistoryBufferX = new float[length];	
        memset(mHistoryBufferX, 0, length*sizeof(float)); 
		mHistoryBufferY = new float[length];	
        memset(mHistoryBufferY, 0, length*sizeof(float));		
	}
	return true;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/**
 *	Filters mouse motion.
 *	\param		delta_mouse_x	[in/out] delta mouse position
 *	\param		delta_mouse_y	[in/out] delta mouse position
 */
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void MouseFilter::Apply(NxF32& delta_mouse_x, NxF32& delta_mouse_y)
{
	// Checkings
	if(!mHistoryBufferX || !mHistoryBufferY)	return;

	// Shift the buffer around. If you want performance from this, be sure
	// to use a circular buffer than these slow memmove()s.
	memmove(mHistoryBufferX+1, mHistoryBufferX, (mHistoryBufferLength-1)*sizeof(NxF32));
	memmove(mHistoryBufferY+1, mHistoryBufferY, (mHistoryBufferLength-1)*sizeof(NxF32));

	// Put the current values at the front of the history buffer
	*mHistoryBufferX = delta_mouse_x;
	*mHistoryBufferY = delta_mouse_y;

	// Filter the mouse
	NxF32 CurAverageX	= 0.0f;
	NxF32 CurAverageY	= 0.0f;
	NxF32 AverageTot	= 0.0f;
	NxF32 CurrentWeight	= 1.0f;
	for(NxU32 i=0;i<mHistoryBufferLength;i++)
	{
		CurAverageX += mHistoryBufferX[i] * CurrentWeight;
		CurAverageY += mHistoryBufferY[i] * CurrentWeight;

		// Note! Our total is also weighted
		AverageTot += 1.0f * CurrentWeight;

		// The weight for the next entry in the history buffer
		CurrentWeight *= mWeightModifier;
	}

	// Calculate the final weighted value
	delta_mouse_x = CurAverageX / AverageTot;
	delta_mouse_y = CurAverageY / AverageTot;
}
